Type TMidiFile
	Field tracks:TMidiTrack[]
	
	Field formatType:Short
	Field numberOfTracks:Short
	Field timeDivision:Short
	Field ticksPerBeat:Short
	Field framesPerSecond:Byte
	Field ticksPerFrame:Byte
	
	Field tempo:Double
	Field _timePerTick:Double		' time per tick in milliseconds
	Const TIMEDIVISION_TPB:Int = 0
	Const TIMEDIVISION_FPS:Int = 1
	
	Field _tick:Int	' song position
	Field _time:Double	' song position in milliseconds
	
	' load midi file
	Function LoadFile:TMidiFile(pFileName:String)
		Local midi:TMidiFile = New TMidiFile
		' open stream for reading
		Local file:TStream = BigEndianStream(ReadFile(pFileName))
		If Not file Then Throw "Unable to load file "+pFileName
		' read midi header
		If file.ReadInt()<>$4D546864 Then Throw pFileName+" is not a valid midi file"
		If file.ReadInt()<>6 Then Throw pFileName+" is corrupt"
		midi.formatType = file.ReadShort()
		midi.numberOfTracks = file.ReadShort()
		midi.timeDivision = file.ReadShort()
		' set up time base
		If midi.timeDivision & $8000 = 0	' ticks per beat
			midi.ticksPerBeat = midi.timeDivision & $7FFF
			midi.timeDivision = TIMEDIVISION_TPB
		Else	' frames per second
			midi.framesPerSecond = (midi.timeDivision & $7F00) Shl 8
			midi.ticksPerFrame = midi.timeDivision & $FF
			midi.timeDivision = TIMEDIVISION_FPS
		EndIf
		' set initial tempo
		midi.SetTempo()
		' read tracks
		For Local t:Int = 0 Until midi.numberOfTracks
			midi.tracks :+ [TMidiTrack._LoadTrack(file,midi)]
		Next
		Return midi
	End Function
	
	Method SetTempo(pTempo:Double=120)
		tempo = pTempo
		If timeDivision = TIMEDIVISION_TPB
			_timePerTick = 60000:Double / Double(tempo *ticksPerBeat)
		Else
			' special case: 29fps = 29.97fps
			_timePerTick = 1000:Double / Double((framesPerSecond+0.97*(framesPerSecond=29)) *ticksPerFrame)
		EndIf
	End Method
	
	' play certain ticks
	Method Play:TMidiTrackEvent[](pTicks:Int)
		Local trackEvents:TMidiTrackEvent[]
		For Local i:Int = 0 Until pTicks
			For Local t:Int = 0 Until tracks.length
				Local events:TMidiEvent[] = tracks[t]._Play(1)
				If events
					Local trackEvent:TMidiTrackEvent = New TMidiTrackEvent
					trackEvent.track = t
					trackEvent.events = events
					trackEvents :+ [trackEvent]
				EndIf
			Next
			_tick :+ 1
			_time :+ _timePerTick
		Next
		Return trackEvents
	End Method
	
	' if song is still playing
	Method Playing:Int()
		For Local t:TMidiTrack = EachIn tracks
			If t._Playing() Then Return True
		Next
	End Method
	
	' return tick count for given playtime
	Method TicksForPlaytime:Int(pTime:Double)
		Return (pTime -_time) / _timePerTick
	End Method
End Type

Type TMidiTrack
	Field _midiFile:TMidiFile
	Field chunkSize:Int
	Field events:TMidiEvent[]
	
	Field _tick:Int	' track position
	Field _lastTick:Int ' tick, when last event was triggered
	Field _lastEvent:Int ' index of last triggered event
	
	' load track
	Function _LoadTrack:TMidiTrack(pStream:TStream, pMidiFile:TMidiFile)
		Local track:TMidiTrack = New TMidiTrack
		track._midiFile = pMidiFile
		' read track header
		If pStream.ReadInt()<>$4D54726B Then Throw "Midifile is corrupt"
		track.chunkSize = pStream.ReadInt()
		' read event data
		Local spos:Int = pStream.Pos()
		While pStream.Pos()-spos < track.chunkSize
			Local deltaTime:Int = _ReadVariableLength(pStream)
			Local typeVal:Byte = pStream.ReadByte()
			' read channel event
			If typeVal<255
				Local par1:Byte = pStream.ReadByte()
				Local par2:Byte = pStream.ReadByte()
				track.events :+ [TMidiEvent._CreateChannelEvent(deltaTime, typeVal, par1, par2)]
			' read meta event
			Else
				track.events :+ [TMidiEvent._CreateMetaEvent(pStream)]
			EndIf
		Wend
		Return track
	End Function
	
	Function _ReadVariableLength:Int(pStream:TStream)
		Local retVal:Int
		Local v:Byte
		Repeat
			v = pStream.ReadByte()
			retVal :+ (v & $7F)
			' following byte
			If v&$80
				retVal :Shl 7
			Else
				Return retVal
			EndIf
		Forever
	End Function
	
	' play certain ticks of track
	Method _Play:TMidiEvent[](pTicks:Int)
		' return null on track end
		If _lastEvent=events.length Then Return Null
		Local _events:TMidiEvent[]
		For Local i:Int = 0 Until pTicks
			' trigger triggerable events
			While _lastEvent<events.length And _tick=events[_lastEvent].deltaTime+_lastTick
				Local ev:TMidiEvent = events[_lastEvent]
				If ev.eventType=255 And ev.metaType = 81
					Local val:Int = (ev.data[0] Shl 16) | (ev.data[1] Shl 8) | ev.data[2]
					Local tempo:Double = 60000000:Double / Double(val)
					_midiFile.SetTempo(tempo)
				EndIf
				_events :+ [ev]
				_lastEvent :+ 1
				_lastTick = _tick
			Wend
			_tick :+ 1	
		Next
		Return _events
	End Method
	
	' if track is still playing
	Method _Playing:Int()
		If _lastEvent<events.length Return True
	End Method
End Type

Type TMidiTrackEvent
	Field track:Int			' track number
	Field events:TMidiEvent[]	' event
	
	Method ToString:String()
		Local txt:String
		txt :+ "- Track "+track+" -~n"
		For Local e:TMidiEvent = EachIn events
			txt :+ "~t"+e.ToString()+"~n"
		Next
		Return txt
	End Method
End Type

Type TMidiEvent
	Const NOTE_OFF:Int			= $8
	Const NOTE_ON:Int			= $9
	Const NOTE_AFTERTOUCH:Int	= $A
	Const CONTROLLER:Int		= $B
	Const PROGRAM_CHANGE:Int		= $C
	Const CHANNEL_AFTERTOUCH:Int	= $D
	Const PITCH_BEND:Int		= $E
	
	Field deltaTime:Int
	
	' channel event
	Field eventType:Byte
	Field channel:Byte
	Field parameter1:Byte
	Field parameter2:Byte
	Function _CreateChannelEvent:TMidiEvent(pDeltaTime:Int, pEventType:Byte, pPar1:Byte, pPar2:Byte)
		Local event:TMidiEvent = New TMidiEvent 
		event.deltaTime = pDeltaTime
		event.eventType = (pEventType & $F0) Shr 4
		event.channel = (pEventType & $0F)
		event.parameter1 = pPar1
		event.parameter2 = pPar2
		Return event
	End Function
	
	' meta event
	Field metaType:Int
	Field length:Int
	Field data:Byte Ptr
	Function _CreateMetaEvent:TMidiEvent(pStream:TStream)
		Local event:TMidiEvent = New TMidiEvent
		' read meta event data
		event.eventType = 255
		event.metaType = pStream.ReadByte()
		event.length = TMidiTrack._ReadVariableLength(pStream)
		event.data = MemAlloc(event.length)
		pStream.ReadBytes(event.data, event.length)
		Return event
	End Function
	
	Method ToString:String()
		Return "Event~t" +eventType+",~t"+channel+",~t"+parameter1+",~t"+parameter2+",~t"+metaType+",~t"+length
	End Method
End Type
